import inspect
from abc import ABCMeta, abstractmethod

import six

from webpub_manifest_parser.core.parsers import (
    AnyOfParser,
    ArrayParser,
    BooleanParser,
    DateParser,
    DateTimeParser,
    EnumParser,
    IntegerParser,
    LocalizableStringParser,
    NumberParser,
    StringParser,
    StringPatternParser,
    TypeParser,
    URIParser,
    URIReferenceParser,
    URITemplateParser,
    ValueParser,
)
from webpub_manifest_parser.utils import is_string


@six.add_metaclass(ABCMeta)
class HasProperties(object):
    """Interface representing class containing ObjectProperty meta-properties."""

    @abstractmethod
    def get_setting_value(self, setting_name, default_value=None):
        """Return the setting's value.

        :param setting_name: Name of the setting
        :type setting_name: string

        :param default_value: Default value
        :type default_value: Any

        :return: Setting's value
        :rtype: Any
        """
        raise NotImplementedError()

    @abstractmethod
    def set_setting_value(self, setting_name, setting_value):
        """Set the setting's value.

        :param setting_name: Name of the setting
        :type setting_name: string

        :param setting_value: New value of the setting
        :type setting_value: Any
        """
        raise NotImplementedError()


class Property(object):
    """Class representing object property, storing property's metadata and its value."""

    _counter = 0

    def __init__(self, key, required, parser, default_value=None):
        """Initialize a new instance of Property class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param parser: Parse used to validate and parse property's value
        :type parser: parsers.ValueParser

        :param default_value: Property's default value
        :type default_value: Any
        """
        if not is_string(key):
            raise ValueError("Argument 'key' must be a string")
        if not isinstance(required, bool):
            raise ValueError("Argument 'required' must be boolean")
        if not isinstance(parser, ValueParser):
            raise ValueError(
                "Argument 'parser' must be an instance of {0}".format(ValueParser)
            )

        self._key = key
        self._required = required
        self._parser = parser
        self._default_value = default_value

    def __get__(self, owner_instance, owner_type):
        """Return the property's value.

        :param owner_instance: Instance of the owner, class having instance of ObjectProperty as an attribute
        :type owner_instance: Optional[HasProperties]

        :param owner_type: Owner's class
        :type owner_type: Optional[Type]

        :return: ObjectProperty instance (when called via a static method) or
            the setting's value (when called via an instance method)
        :rtype: Union[ObjectProperty, Any]
        """
        # If owner_instance is empty, it means that this method was called via a static method
        # In this case we need to return the metadata instance itself
        if owner_instance is None:
            return self

        if not isinstance(owner_instance, HasProperties):
            raise ValueError(
                "Argument 'owner_instance' must be an instance of {0}".format(
                    HasProperties
                )
            )

        return owner_instance.get_setting_value(self._key, self._default_value)

    def __set__(self, owner_instance, value):
        """Set the property's value.

        :param owner_instance: Instance of the owner, class having instance of ObjectProperty as an attribute
        :type owner_instance: Optional[HasProperties]

        :param value: New setting's value
        :type value: Any
        """
        if not isinstance(owner_instance, HasProperties):
            raise ValueError(
                "Argument 'owner_instance' must be an instance of {0}".format(
                    HasProperties
                )
            )

        return owner_instance.set_setting_value(self._key, value)

    def __repr__(self):
        """Return a string representation of the object.

        :return: String representation
        :rtype: str
        """
        return (
            u"<Property(key={0}, required={1}, parser={2}, default_value={3})>".format(
                self.key, self.required, self.parser, self.default_value
            )
        )

    @property
    def key(self):
        """Return the property's key.

        :return: Property's key
        :rtype: str
        """
        return self._key

    @property
    def required(self):
        """Return a boolean value indicating whether this property is required or not.

        :return: Boolean value indicating whether this property is required or not.
        :rtype: bool
        """
        return self._required

    @property
    def parser(self):
        """Return the parser used to validate and parse this property's value.

        :return: Parser used to validate and parse this property's value.
        :rtype: parsers.Parser
        """
        return self._parser

    @property
    def default_value(self):
        """Return the property's default value.

        :return: Property's default value.
        :rtype: bool
        """
        return self._default_value


class PropertiesGrouping(HasProperties):
    """Group of properties."""

    def __init__(self):
        """Initialize a new instance of PropertiesGrouping class."""
        self._values = {}

    def get_setting_value(self, setting_name, default_value=None):
        """Return the setting's value.

        :param setting_name: Setting's name
        :type setting_name: str

        :param default_value: Setting's default value
        :type default_value: Any
        """
        return self._values.get(setting_name, default_value)

    def set_setting_value(self, setting_name, setting_value):
        """Set the setting's value.

        :param setting_name: Setting's name
        :type setting_name: str

        :param setting_value: New setting's value
        :type setting_value: Any
        """
        self._values[setting_name] = setting_value

    @staticmethod
    def get_class_properties(klass):
        """Return a list of 2-tuples containing information ConfigurationMetadata properties in the specified class.

        :param klass: Class
        :type klass: type

        :return: List of 2-tuples containing information ConfigurationMetadata properties in the specified class
        :rtype: List[Tuple[string, ConfigurationMetadata]]
        """
        members = inspect.getmembers(klass, lambda member: isinstance(member, Property))

        return members

    @staticmethod
    def get_required_class_properties(klass):
        """Return a list of 2-tuples containing information about required ConfigurationMetadata properties.

        :param klass: Class
        :type klass: type

        :return: List of 2-tuples containing information ConfigurationMetadata properties in the specified class
        :rtype: List[Tuple[string, ConfigurationMetadata]]
        """
        class_properties = PropertiesGrouping.get_class_properties(klass)
        required_class_properties = [
            (class_property_name, class_property)
            for (class_property_name, class_property) in class_properties
            if class_property.required
        ]

        return required_class_properties


@six.add_metaclass(ABCMeta)
class ParsableProperty(Property):
    """Base class for all property classes having predefined parsers."""

    PARSER = object()

    def __init__(self, key, required, default_value=None):
        """Initialize a new instance of ParsableProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param default_value: Property's default value
        :type default_value: Any
        """
        super(ParsableProperty, self).__init__(
            key, required, self.PARSER, default_value
        )


class IntegerProperty(Property):
    """Property allowing only integer values."""

    def __init__(  # pylint: disable=R0913
        self,
        key,
        required,
        minimum=None,
        exclusive_minimum=None,
        maximum=None,
        exclusive_maximum=None,
        default_value=None,
    ):
        """Initialize a new instance of IntegerProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param minimum: Minimum value
        :type minimum: Numeric

        :param exclusive_minimum: Exclusive minimum value
        :type exclusive_minimum: Numeric

        :param maximum: Maximum value
        :type maximum: Numeric

        :param exclusive_maximum: Exclusive maximum value
        :type exclusive_maximum: Numeric

        :param default_value: Property's default value
        :type default_value: Any
        """
        super(IntegerProperty, self).__init__(
            key,
            required,
            IntegerParser(minimum, exclusive_minimum, maximum, exclusive_maximum),
            default_value,
        )


class NumberProperty(Property):
    """Property allowing only float values."""

    def __init__(  # pylint: disable=R0913
        self,
        key,
        required,
        minimum=None,
        exclusive_minimum=None,
        maximum=None,
        exclusive_maximum=None,
        default_value=None,
    ):
        """Initialize a new instance of NumberProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param minimum: Minimum value
        :type minimum: Numeric

        :param exclusive_minimum: Exclusive minimum value
        :type exclusive_minimum: Numeric

        :param maximum: Maximum value
        :type maximum: Numeric

        :param exclusive_maximum: Exclusive maximum value
        :type exclusive_maximum: Numeric

        :param default_value: Property's default value
        :type default_value: Any
        """
        super(NumberProperty, self).__init__(
            key,
            required,
            NumberParser(minimum, exclusive_minimum, maximum, exclusive_maximum),
            default_value,
        )


class BooleanProperty(ParsableProperty):
    """Property allowing only boolean values."""

    PARSER = BooleanParser()


class StringProperty(ParsableProperty):
    """Property allowing only string values."""

    PARSER = StringParser()


class EnumProperty(Property):
    """Property allowing only specific string values."""

    def __init__(self, key, required, items, default_value=None):
        """Initialize a new instance of EnumProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param items: Enumeration values
        :type items: List[str]

        :param default_value: Property's default value
        :type default_value: Any
        """
        if not isinstance(items, list):
            raise ValueError("Argument 'items' must be an instance of {0}".format(list))

        super(EnumProperty, self).__init__(
            key, required, EnumParser(items), default_value
        )


class URIProperty(ParsableProperty):
    """Property allowing only URI values."""

    PARSER = URIParser()


class URITemplateProperty(ParsableProperty):
    """Property allowing only URI templates."""

    PARSER = URITemplateParser()


class URIReferenceProperty(ParsableProperty):
    """Property allowing only URI-reference values."""

    PARSER = URIReferenceParser()


class DateProperty(ParsableProperty):
    """Property allowing only date values."""

    PARSER = DateParser()


class DateTimeProperty(ParsableProperty):
    """Property allowing only date & time values."""

    PARSER = DateTimeParser()


class DateOrTimeProperty(ParsableProperty):
    """Property allowing date or date & time values."""

    PARSER = AnyOfParser([DateParser(), DateTimeParser()])


class BaseArrayProperty(Property):
    """Property containing a list of items."""

    def __init__(
        self, key, required, parser, list_type=list, default_value=None
    ):  # pylint: disable=R0913
        """Initialize a new instance of ListProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param parser: Parse used to validate and parse property's value
        :type parser: parsers.ValueParser

        :param list_type: Class of a collection used to hold the items
        :type list_type: Type

        :param default_value: Property's default value
        :type default_value: Any
        """
        if not issubclass(list_type, list):
            raise ValueError(
                "Argument 'list_type' must be a subclass of {0}".format(list)
            )

        super(BaseArrayProperty, self).__init__(key, required, parser, default_value)

        self._list_type = list_type

    def __set__(self, owner_instance, value):
        """Set the property's value.

        :param owner_instance: Instance of the owner, class having instance of ObjectProperty as an attribute
        :type owner_instance: Optional[HasProperties]

        :param value: New setting's value
        :type value: Any
        """
        if value is not None and not isinstance(value, self._list_type):
            raise ValueError(
                "Value must be a subclass of {0} class".format(self._list_type)
            )

        super(BaseArrayProperty, self).__set__(owner_instance, value)

    @property
    def list_type(self):
        """Return the class of a collection used to hold the items.

        :return: Class of a collection used to hold the items
        :rtype: Type
        """
        return self._list_type


class ArrayProperty(BaseArrayProperty):
    """Property containing an array of items."""

    def __init__(  # pylint: disable=R0913
        self,
        key,
        required,
        item_parser,
        unique_items=False,
        list_type=list,
        default_value=None,
    ):
        """Initialize a new instance of ArrayProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param item_parser: Parser for array items
        :type item_parser: ValueParser

        :param unique_items: Boolean value indicating whether the array should contain unique items
        :type unique_items: bool

        :param list_type: Class of a collection used to hold the items
        :type list_type: Type

        :param default_value: Property's default value
        :type default_value: Any
        """
        if not isinstance(item_parser, ValueParser):
            raise ValueError(
                "Argument 'item_parser' must be an instance of {0}".format(ValueParser)
            )
        if not isinstance(unique_items, bool):
            raise ValueError("Argument 'unique_items' must be boolean")

        super(ArrayProperty, self).__init__(
            key,
            required,
            ArrayParser(item_parser, unique_items),
            list_type,
            default_value,
        )


class ArrayOfStringsProperty(BaseArrayProperty):
    """Property allowing either a string or array of strings as its values."""

    def __init__(  # pylint: disable=R0913
        self, key, required, unique_items=False, list_type=list, default_value=None
    ):
        """Initialize a new instance of ArrayOfStringsProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param unique_items: Boolean value indicating whether the array should contain unique items
        :type unique_items: bool

        :param list_type: Class of a collection used to hold the items
        :type list_type: Type

        :param default_value: Property's default value
        :type default_value: Any
        """
        if not isinstance(unique_items, bool):
            raise ValueError("Argument 'unique_items' must be boolean")

        super(ArrayOfStringsProperty, self).__init__(
            key,
            required,
            AnyOfParser([ArrayParser(StringParser(), unique_items), StringParser()]),
            list_type,
            default_value,
        )


class ListOfLanguagesProperty(BaseArrayProperty):
    """Property allowing localizable strings.

    For example:
        - "en"
        - [
            "eng",
            "fre"
          ]
    """

    def __init__(self, key, required):  # pylint: disable=R0913
        """Initialize a new instance of ListOfLanguagesProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool
        """
        super(ListOfLanguagesProperty, self).__init__(
            key,
            required,
            AnyOfParser(
                [
                    StringPatternParser(LocalizableStringParser.LANGUAGE_PATTERN),
                    ArrayParser(
                        StringPatternParser(LocalizableStringParser.LANGUAGE_PATTERN)
                    ),
                ]
            ),
            list,
            [],
        )


class LocalizableStringProperty(ParsableProperty):
    """Property allowing either only string/localizable string values.

    For example:
    - "plain string"
    - {
        "eng": "Hello",
        "esp": "Hola"
      }
    """

    PARSER = AnyOfParser([LocalizableStringParser(), StringParser()])


class TypeProperty(Property):
    """Property allowing only specific values of a specific Type."""

    def __init__(self, key, required, nested_type, default_value=None):
        """Initialize a new instance of TypeProperty class.

        :param key: Property's key
        :type key: str

        :param required: Boolean value indicating whether the property is required or not
        :type required: bool

        :param nested_type: Value's type
        :type nested_type: Type

        :param default_value: Property's default value
        :type default_value: Any
        """
        super(TypeProperty, self).__init__(
            key, required, TypeParser(nested_type), default_value
        )



import io
import json
import logging
from abc import ABCMeta, abstractmethod
from io import StringIO

import requests  # noqa: I201
import six  # noqa: I201

from webpub_manifest_parser.core.analyzer import AnalyzerContext


class ManifestParserResult(AnalyzerContext):
    """Class containing the result of the semantic analysis: root AST node and a list of found errors."""

    def __init__(self, root):
        """Initialize a new instance of DocumentParserResult class.

        :param root: Root AST node
        :type root: webpub_manifest_parser.core.ast.Node
        """
        super(ManifestParserResult, self).__init__()

        self._root = root

    @property
    def root(self):
        """Return the root AST node.

        :return: Root AST node
        :rtype: webpub_manifest_parser.core.ast.Manifestlike
        """
        return self._root


class ManifestParser(object):
    """Base class for RWPM-compatible parsers."""

    def __init__(self, syntax_analyzer, semantic_analyzer):
        """Initialize a new instance of ManifestParser class.

        :param syntax_analyzer: Syntax analyzer
        :type syntax_analyzer: syntax.SyntaxAnalyzer

        :param semantic_analyzer: Semantic analyser
        :type semantic_analyzer: semantic.SemanticAnalyzer
        """
        self._syntax_analyzer = syntax_analyzer
        self._semantic_analyzer = semantic_analyzer

        self._logger = logging.getLogger(__name__)

    def _parse(self, manifest_json):
        """Parse the JSON object containing an RWPM-compatible manifest.

        :param manifest_json: JSON object containing an RWPM-compatible manifest
        :type manifest_json: Dict

        :return: Parser result containing an AST object and a list of syntax and semantic errors
        :rtype: ManifestParserResult
        """
        manifest = self._syntax_analyzer.analyze(manifest_json)
        result = ManifestParserResult(manifest)

        manifest.accept(self._semantic_analyzer)

        result.errors.extend(self._syntax_analyzer.context.errors)
        result.errors.extend(self._semantic_analyzer.context.errors)

        return result

    def parse_file(self, input_file_path, encoding="utf-8"):
        """Parse the input file and return a validated AST object.

        :param input_file_path: Full path to the file containing RWPM-compatible document
        :type input_file_path: str

        :param encoding: Input file's encoding
        :type encoding: str

        :return: Parser result
        :rtype: ManifestParserResult
        """
        with io.open(input_file_path, "r", encoding=encoding) as input_file:
            manifest_json = self.get_manifest_json(input_file)

            return self._parse(manifest_json)

    def parse_stream(self, input_stream):
        """Parse the input file and return the result: root AST object and a list of errors.

        :param input_stream: Full path to the file containing RWPM-compatible document
        :type input_stream: six.StringIO

        :return: Parser result
        :rtype: ManifestParserResult
        """
        manifest_json = self.get_manifest_json(input_stream)

        return self._parse(manifest_json)

    def parse_url(
        self, url, encoding="utf-8", params=None, auth=None, proxies=None
    ):  # pylint: disable=too-many-arguments
        """Fetch the content pointed by the URL, parse it and return the result: root AST object and a list of errors.

        :param url: URL pointing to the RWPM-compatible document
        :type url: str

        :param encoding: Input file's encoding
        :type encoding: str

        :param params: Dictionary containing query parameters
        :type params: Dict

        :param auth: Authentication information
        :type auth: requests.auth.AuthBase

        :param proxies: Dictionary containing proxy information
        :type proxies: Dict

        :return: Parser result
        :rtype: ManifestParserResult
        """
        response = requests.get(url, params=params, auth=auth, proxies=proxies)
        input_stream = StringIO(six.text_type(response.content, encoding))
        manifest_json = self.get_manifest_json(input_stream)

        return self._parse(manifest_json)

    def parse_json(self, manifest_json):
        """Parse the JSON document with an RWPM-compatible manifest and return the result.

        :param manifest_json: JSON document with an RWPM-compatible manifest
        :type manifest_json: Dict

        :return: Parser result containing the root AST object and a list of errors
        :rtype: ManifestParserResult
        """
        return self._parse(manifest_json)

    @staticmethod
    def get_manifest_json(input_stream):
        """Parse the input stream into a JSON document containing an RWPM-compatible manifest.

        :param input_stream: Input stream containing JSON document with an RWPM-compatible manifest
        :type input_stream: Union[six.StringIO, six.BinaryIO]

        :return: JSON document containing an RWPM-compatible manifest
        :rtype: Dict
        """
        logging.debug("Started parsing input stream into a JSON document")

        input_stream_content = input_stream.read()
        input_stream_content = input_stream_content.strip()
        manifest_json = json.loads(input_stream_content)

        logging.debug("Finished parsing input stream into a JSON document")

        return manifest_json


@six.add_metaclass(ABCMeta)
class ManifestParserFactory(object):
    """Base class for factories creating parsers for particular RWPM-compatible standards (for example, OPDS 2.0)."""

    @abstractmethod
    def create(self):
        """Create a new Parser instance.

        :return: Parser instance
        :rtype: ManifestParser
        """
        raise NotImplementedError()



from webpub_manifest_parser.core.registry import CollectionRole, LinkRelation, MediaType
from webpub_manifest_parser.rwpm.registry import (
    RWPMCollectionRolesRegistry,
    RWPMLinkRelationsRegistry,
    RWPMMediaTypesRegistry,
)


class OPDS2MediaTypesRegistry(RWPMMediaTypesRegistry):
    """Registry containing OPDS 2.0 media types."""

    # https://drafts.opds.io/opds-2.0.html#overview
    OPDS_FEED = MediaType(key="application/opds+json")

    # https://drafts.opds.io/opds-2.0.html#41-opds-publication
    OPDS_PUBLICATION = MediaType(key="application/opds-publication+json")

    OPDS_MEDIA_TYPES = [OPDS_FEED, OPDS_PUBLICATION]

    def __init__(self):
        """Initialize a new instance of OPDS2MediaTypesRegistry class."""
        super(OPDS2MediaTypesRegistry, self).__init__()

        self._add_items(self.OPDS_MEDIA_TYPES)


class OPDS2LinkRelationsRegistry(RWPMLinkRelationsRegistry):
    """Registry containing OPDS 2.0 link relations."""

    ACQUISITION = LinkRelation(key="http://opds-spec.org/acquisition")
    OPEN_ACCESS = LinkRelation(key="http://opds-spec.org/acquisition/open-access")
    BORROW = LinkRelation(key="http://opds-spec.org/acquisition/borrow")
    BUY = LinkRelation(key="http://opds-spec.org/acquisition/buy")
    SAMPLE = LinkRelation(key="http://opds-spec.org/acquisition/sample")
    PREVIEW = LinkRelation(key="preview")
    SUBSCRIBE = LinkRelation(key="http://opds-spec.org/acquisition/subscribe")

    CORE_LINK_RELATIONS = [
        ACQUISITION,
        OPEN_ACCESS,
        BORROW,
        BUY,
        SAMPLE,
        PREVIEW,
        SUBSCRIBE,
    ]

    def __init__(self):
        """Initialize a new instance of OPDS2LinkRelationsRegistry class."""
        super(OPDS2LinkRelationsRegistry, self).__init__()

        self._add_items(self.CORE_LINK_RELATIONS)


class OPDS2CollectionRolesRegistry(RWPMCollectionRolesRegistry):
    """Registry containing ODPS 2.0 collection roles."""

    NAVIGATION = CollectionRole(key="navigation", compact=True, required=False)
    PUBLICATIONS = CollectionRole(key="publications", compact=False, required=False)
    IMAGES = CollectionRole(key="images", compact=True, required=False)
    FACETS = CollectionRole(key="facets", compact=False, required=False)
    GROUPS = CollectionRole(key="groups", compact=False, required=False)

    OPDS_2_0_ROLES = [NAVIGATION, PUBLICATIONS, IMAGES, FACETS, GROUPS]

    def __init__(self):
        """Initialize a new instance of OPDS2CollectionRolesRegistry class."""
        super(OPDS2CollectionRolesRegistry, self).__init__()

        self._add_items(self.OPDS_2_0_ROLES)



import logging
from functools import partial

from multipledispatch import dispatch

from webpub_manifest_parser.core.ast import (
    Collection,
    CollectionList,
    CompactCollection,
    Link,
    LinkList,
    Manifestlike,
    Metadata,
)
from webpub_manifest_parser.core.semantic import SemanticAnalyzer, SemanticAnalyzerError
from webpub_manifest_parser.opds2.ast import (
    OPDS2FeedMetadata,
    OPDS2Group,
    OPDS2Navigation,
    OPDS2Publication,
)
from webpub_manifest_parser.opds2.registry import OPDS2LinkRelationsRegistry
from webpub_manifest_parser.utils import cast, encode

MISSING_REQUIRED_FEED_SUB_COLLECTIONS = partial(
    SemanticAnalyzerError,
    message="OPDS 2.0 feed must contain one of the following sub-collections: publications, navigation, groups",
)

MISSING_NAVIGATION_LINK_TITLE_ERROR = partial(
    SemanticAnalyzerError, message="OPDS 2.0 navigation link must contain a title"
)

MISSING_ACQUISITION_LINK = partial(
    SemanticAnalyzerError,
    message="OPDS 2.0 publication must contain at least one acquisition link",
)

WRONG_GROUP_STRUCTURE = partial(
    SemanticAnalyzerError,
    message="OPDS 2.0 group must contain either a single navigation collection or a single publications collection",
)


class OPDS2SemanticAnalyzer(SemanticAnalyzer):
    """OPDS 2.0 semantic analyzer."""

    def __init__(
        self, media_types_registry, link_relations_registry, collection_roles_registry
    ):
        """Initialize a new instance of OPDS2SemanticAnalyzer class.

        :param media_types_registry: Media types registry
        :type media_types_registry: python_rwpm_parser.registry.Registry

        :param link_relations_registry: Link relations registry
        :type link_relations_registry: python_rwpm_parser.registry.Registry

        :param collection_roles_registry: Collections roles registry
        :type collection_roles_registry: python_rwpm_parser.registry.Registry
        """
        super(OPDS2SemanticAnalyzer, self).__init__(
            media_types_registry, link_relations_registry, collection_roles_registry
        )

        self._logger = logging.getLogger(__name__)

    @dispatch(Manifestlike)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the manifest node.

        :param node: Manifest's metadata
        :type node: OPDS2Feed
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        super(OPDS2SemanticAnalyzer, self).visit(node)

        if (
            node.publications is None
            and node.navigation is None
            and node.groups is None
        ):
            with self._record_errors():
                raise MISSING_REQUIRED_FEED_SUB_COLLECTIONS(
                    node=node, node_property=None
                )

        if node.publications is not None:
            with self._record_errors():
                node.publications.accept(self)
        if node.navigation is not None:
            with self._record_errors():
                node.navigation.accept(self)
        if node.groups is not None:
            with self._record_errors():
                node.groups.accept(self)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(OPDS2FeedMetadata)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the feed's metadata.

        :param node: Feed's metadata
        :type node: OPDS2FeedMetadata
        """

    @dispatch(Metadata)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the publication's metadata.

        :param node: Publication's metadata
        :type node: Metadata
        """
        super(OPDS2SemanticAnalyzer, self).visit(node)

    @dispatch(LinkList)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the list of links.

        :param node: Manifest's metadata
        :type node: LinkList
        """
        super(OPDS2SemanticAnalyzer, self).visit(node)

    @dispatch(Link)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the link node.

        :param node: Link node
        :type node: Link
        """
        super(OPDS2SemanticAnalyzer, self).visit(node)

    @dispatch(CollectionList)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the list of sub-collections.

        :param node: CollectionList node
        :type node: CollectionList
        """
        super(OPDS2SemanticAnalyzer, self).visit(node)

    @dispatch(OPDS2Publication)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the OPDS 2.0 publication.

        :param node: OPDS 2.0 publication
        :type node: OPDS2Publication
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        super(OPDS2SemanticAnalyzer, self).visit(node)

        acquisition_links = [
            OPDS2LinkRelationsRegistry.PREVIEW.key,
            OPDS2LinkRelationsRegistry.ACQUISITION.key,
            OPDS2LinkRelationsRegistry.BUY.key,
            OPDS2LinkRelationsRegistry.OPEN_ACCESS.key,
            OPDS2LinkRelationsRegistry.BORROW.key,
            OPDS2LinkRelationsRegistry.SAMPLE.key,
            OPDS2LinkRelationsRegistry.SUBSCRIBE.key,
        ]

        for link in node.links:
            if link.rels is not None and set(acquisition_links) & set(link.rels):
                break
        else:
            with self._record_errors():
                raise MISSING_ACQUISITION_LINK(node=node, node_property=None)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(OPDS2Group)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the OPDS 2.0 group.

        :param node: OPDS 2.0 group
        :type node: OPDS2Group
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        # FIXME: It seems that group definition relaxes requirements for having metadata
        # It means we have to override default behaviour
        # super(OPDS2SemanticAnalyzer, self).visit(node)

        if node.metadata:
            with self._record_errors():
                node.metadata.accept(self)

        if node.publications and node.navigation:
            with self._record_errors():
                raise WRONG_GROUP_STRUCTURE(node=node, node_property=None)

        if node.publications:
            with self._record_errors():
                node.publications.accept(self)
        if node.navigation:
            with self._record_errors():
                node.navigation.accept(self)
        if node.links:
            with self._record_errors():
                node.links.accept(self)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(OPDS2Navigation)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the OPDS 2.0 navigation.

        :param node: OPDS 2.0 navigation
        :type node: OPDS2Navigation
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        with self._record_errors():
            self.visit(cast(node, CompactCollection))

        for link in node.links:
            if link.title is None:
                with self._record_errors():
                    raise MISSING_NAVIGATION_LINK_TITLE_ERROR(
                        node=link, node_property=Link.title
                    )

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(CompactCollection)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the compact collection node.

        :param node: Collection node
        :type node: CompactCollection
        """
        super(OPDS2SemanticAnalyzer, self).visit(node)

    @dispatch(Collection)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the collection node.

        :param node: Collection node
        :type node: Collection
        """
        super(OPDS2SemanticAnalyzer, self).visit(node)



from webpub_manifest_parser.core.registry import (
    CollectionRole,
    LinkRelation,
    LinkRelationsRegistry,
    MediaType,
    Registry,
)


class RWPMCollectionRolesRegistry(Registry):
    """Registry containing collection roles defined in the RWPM spec."""

    READING_ORDER = CollectionRole(key="readingOrder", compact=True, required=True)
    RESOURCES = CollectionRole(key="resources", compact=True, required=False)
    TOC = CollectionRole(key="toc", compact=True, required=False)

    GUIDED = CollectionRole(key="guided", compact=True, required=False)
    LANDMARKS = CollectionRole(key="landmarks", compact=True, required=False)
    LOA = CollectionRole(key="loa", compact=True, required=False)
    LOI = CollectionRole(key="loi", compact=True, required=False)
    LOT = CollectionRole(key="lot", compact=True, required=False)
    LOV = CollectionRole(key="lov", compact=True, required=False)
    PAGE_LIST = CollectionRole(key="pageList", compact=True, required=False)

    CORE_ROLES = [READING_ORDER, RESOURCES, TOC]

    EXTENSIONS = [GUIDED, LANDMARKS, LOA, LOI, LOT, LOV, PAGE_LIST]

    def __init__(self):
        """Initialize a new instance of RWPMCollectionRolesRegistry class."""
        super(RWPMCollectionRolesRegistry, self).__init__(
            self.CORE_ROLES + self.EXTENSIONS
        )


class RWPMLinkRelationsRegistry(LinkRelationsRegistry):
    """Registry containing link relations mentioned in the RWPM spec."""

    ALTERNATE = LinkRelation(key="alternate")
    CONTENTS = LinkRelation(key="contents")
    COVER = LinkRelation(key="cover")
    MANIFEST = LinkRelation(key="manifest")
    SEARCH = LinkRelation(key="search")

    CORE_LINK_RELATIONS = [
        ALTERNATE,
        CONTENTS,
        COVER,
        MANIFEST,
        SEARCH,
        LinkRelationsRegistry.SELF,
    ]

    def __init__(self):
        """Initialize a new instance of RWPMLinkRelationsRegistry class."""
        super(RWPMLinkRelationsRegistry, self).__init__(self.CORE_LINK_RELATIONS)


class RWPMMediaTypesRegistry(Registry):
    """Registry containing media types mentioned in the RWPM spec."""

    # https://github.com/readium/webpub-manifest#4-media-type
    MANIFEST = MediaType(key="application/webpub+json")

    # https://github.com/readium/webpub-manifest#6-table-of-contents
    HTML = MediaType(key="text/html")
    CSS = MediaType(key="text/css")

    # https://github.com/readium/webpub-manifest#7-cover
    JPEG = MediaType(key="image/jpeg")
    PNG = MediaType(key="image/png")
    GIF = MediaType(key="image/gif")
    WEBP = MediaType(key="image/webp")
    SVG = MediaType(key="image/svg")
    SVG_XML = MediaType(key="image/svg+xml")

    # https://github.com/readium/webpub-manifest#9-package
    WEB_PUBLICATION_PACKAGE = MediaType("application/webpub+zip")
    EPUB_PUBLICATION_PACKAGE = MediaType("application/epub+zip")

    CORE_TYPES = [
        MANIFEST,
        HTML,
        CSS,
        JPEG,
        PNG,
        GIF,
        WEBP,
        SVG,
        SVG_XML,
        WEB_PUBLICATION_PACKAGE,
        EPUB_PUBLICATION_PACKAGE,
    ]

    def __init__(self):
        """Initialize a new instance of RWPMMediaTypesRegistry class."""
        super(RWPMMediaTypesRegistry, self).__init__(self.CORE_TYPES)



import logging
from functools import partial

from multipledispatch import dispatch

from webpub_manifest_parser.core.analyzer import BaseAnalyzer, BaseAnalyzerError
from webpub_manifest_parser.core.ast import (
    Collection,
    CollectionList,
    CompactCollection,
    Link,
    LinkList,
    Manifestlike,
    Metadata,
    Visitor,
)
from webpub_manifest_parser.core.parsers import (
    URIParser,
    URIReferenceParser,
    ValueParserError,
)
from webpub_manifest_parser.core.registry import LinkRelationsRegistry
from webpub_manifest_parser.utils import encode, first_or_default


class SemanticAnalyzerError(BaseAnalyzerError):
    """Exception raised in the case of semantic errors."""

    def __init__(self, node, node_property=None, message=None, inner_exception=None):
        """Initialize a new instance of ManifestSemanticError class.

        :param node: RWPM node
        :type node: webpub_manifest_parser.core.ast.Node

        :param node_property: AST node's property associated with the error
        :type node_property: Optional[webpub_manifest_parser.core.properties.Property]

        :param message: Parameterized string containing description of the error occurred
            and the placeholder for the feed's identifier
        :type message: Optional[str]

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        message = self._format_message(node, node_property, message, inner_exception)

        super(SemanticAnalyzerError, self).__init__(
            node, node_property, message, inner_exception
        )

    def _format_message(
        self,
        node,  # pylint: disable=unused-argument
        node_property=None,  # pylint: disable=unused-argument
        message=None,
        inner_exception=None,  # pylint: disable=unused-argument
    ):
        """Format the error message.

        :param node: RWPM node
        :type node: webpub_manifest_parser.core.ast.Node

        :param node_property: AST node's property associated with the error
        :type node_property: Optional[webpub_manifest_parser.core.properties.Property]

        :param message: Parameterized string containing description of the error occurred
        :type message: Optional[str]

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        return message


class ManifestSemanticError(SemanticAnalyzerError):
    """Base class for semantic errors related to RWPM-like manifests."""

    def _format_message(
        self, node, node_property=None, message=None, inner_exception=None
    ):
        """Format the error message.

        :param node: RWPM-like manifest
        :type node: webpub_manifest_parser.core.ast.Manifestlike

        :param node_property: AST node's property associated with the error
        :type node_property: Optional[webpub_manifest_parser.core.properties.Property]

        :param message: Parameterized string containing description of the error occurred
            and the placeholder for the feed's identifier
        :type message: Optional[str]

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        if not isinstance(node, Manifestlike):
            raise ValueError(
                "Argument 'node' must be an instance of {0}".format(Manifestlike)
            )

        if node.metadata:
            if node.metadata.title:
                message = message.format(node.metadata.title)
            elif node.metadata.identifier:
                message = message.format(node.metadata.identifier)

        return message


class LinkSemanticError(SemanticAnalyzerError):
    """Base class for semantic errors related to RWPM-like links."""

    def _format_message(
        self, node, node_property=None, message=None, inner_exception=None
    ):
        """Format the error message.

        :param node: RWPM-like link
        :type node: webpub_manifest_parser.core.ast.Link

        :param node_property: AST node's property associated with the error
        :type node_property: Optional[webpub_manifest_parser.core.properties.Property]

        :param message: Parameterized string containing description of the error occurred
            and the placeholder for the feed's identifier
        :type message: Optional[str]

        :param inner_exception: (Optional) Inner exception
        :type inner_exception: Optional[Exception]
        """
        if not isinstance(node, Link):
            raise ValueError("Argument 'node' must be an instance of {0}".format(Link))

        message = message.format(node.href)

        return message


MANIFEST_LINK_MISSING_REL_PROPERTY_ERROR = partial(
    LinkSemanticError,
    message="Manifest link '{0}' does not have a required 'rel' property",
)

MANIFEST_MISSING_SELF_LINK_ERROR = partial(
    ManifestSemanticError, message="Manifest '{0}' does not have a required 'self' link"
)

MANIFEST_SELF_LINK_WRONG_HREF_FORMAT_ERROR = partial(
    LinkSemanticError,
    message="Manifest 'self' link's href {0} is incorrect: "
    "it must be an absolute URI to the canonical location of the manifest",
)


class CollectionWrongFormatError(SemanticAnalyzerError):
    """Exception raised in the case when collection's format (compact, full) doesn't not conform with its role."""

    def __init__(self, collection, inner_exception=None):
        """Initialize a new instance of CollectionWrongFormat class.

        :param collection: Collection with a wrong format
        :type collection: python_rwpm_parser.ast.Collection

        :param inner_exception: (Optional) inner exception
        :type inner_exception: Optional[Exception]
        """
        message = "Collection {0} must be {1} but it is not".format(
            collection.role.key, "compact" if collection.role.compact else "full"
        )

        super(CollectionWrongFormatError, self).__init__(
            collection, None, message, inner_exception
        )

        self._collection = collection

    @property
    def collection(self):
        """Return a collection with a wrong format.

        :return: Collection with a wrong format
        :rtype: python_rwpm_parser.ast.Collection
        """
        return self._collection


class SemanticAnalyzer(BaseAnalyzer, Visitor):
    """Visitor performing semantic analysis of the RWPM-compatible documents."""

    def __init__(
        self, media_types_registry, link_relations_registry, collection_roles_registry
    ):
        """Initialize a new instance of SemanticAnalyzer.

        :param media_types_registry: Media types registry
        :type media_types_registry: webpub_manifest_parser.core.registry.Registry

        :param link_relations_registry: Link relations registry
        :type link_relations_registry: webpub_manifest_parser.core.registry.Registry

        :param collection_roles_registry: Collections roles registry
        :type collection_roles_registry: webpub_manifest_parser.core.registry.Registry
        """
        super(SemanticAnalyzer, self).__init__()

        self._media_types_registry = media_types_registry
        self._link_relations_registry = link_relations_registry
        self._collection_roles_registry = collection_roles_registry
        self._logger = logging.getLogger(__name__)

    def _check_manifest_self_link(self, node):
        """Ensure that manifest contains a correctly formatted self link.

        :param node: Manifest-like node
        :type node: Manifestlike
        """
        for link in node.links:
            if not link.rels:
                with self._record_errors():
                    raise MANIFEST_LINK_MISSING_REL_PROPERTY_ERROR(
                        node=link, node_property=Link.rels
                    )

        self_link = first_or_default(
            node.links.get_by_rel(LinkRelationsRegistry.SELF.key)
        )

        if self_link is None:
            raise MANIFEST_MISSING_SELF_LINK_ERROR(node=node, node_property=None)

        parser = URIParser()

        try:
            parser.parse(self_link.href)
        except ValueParserError:
            raise MANIFEST_SELF_LINK_WRONG_HREF_FORMAT_ERROR(
                node=self_link, node_property=Link.href
            )

    @dispatch(Manifestlike)
    def visit(self, node):
        """Perform semantic analysis of the manifest node.

        :param node: Manifest-like node
        :type node: Manifestlike
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        self.context.reset()

        with self._record_errors():
            node.metadata.accept(self)

        with self._record_errors():
            node.links.accept(self)

        with self._record_errors():
            self._check_manifest_self_link(node)

        with self._record_errors():
            node.sub_collections.accept(self)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(Metadata)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the manifest's metadata.

        :param node: Manifest's metadata
        :type node: Metadata
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(LinkList)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the list of links.

        :param node: Manifest's metadata
        :type node: LinkList
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        for link in node:
            with self._record_errors():
                link.accept(self)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(Link)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the link node.

        :param node: Link node
        :type node: Link
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        if not node.templated:
            parser = URIReferenceParser()
            parser.parse(node.href)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(CollectionList)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the list of sub-collections.

        :param node: CollectionList node
        :type node: CollectionList
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        for collection in node:
            with self._record_errors():
                collection.accept(self)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(CompactCollection)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the compact collection node.

        :param node: Collection node
        :type node: CompactCollection
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        with self._record_errors():
            node.links.accept(self)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))

    @dispatch(Collection)  # noqa: F811
    def visit(self, node):  # pylint: disable=E0102
        """Perform semantic analysis of the collection node.

        :param node: Collection node
        :type node: Collection
        """
        self._logger.debug(u"Started processing {0}".format(encode(node)))

        with self._record_errors():
            node.metadata.accept(self)

        with self._record_errors():
            node.links.accept(self)

        with self._record_errors():
            node.sub_collections.accept(self)

        self._logger.debug(u"Finished processing {0}".format(encode(node)))



from webpub_manifest_parser.core import ManifestParser, ManifestParserFactory
from webpub_manifest_parser.rwpm.registry import (
    RWPMCollectionRolesRegistry,
    RWPMLinkRelationsRegistry,
    RWPMMediaTypesRegistry,
)
from webpub_manifest_parser.rwpm.semantic import RWPMSemanticAnalyzer
from webpub_manifest_parser.rwpm.syntax import RWPMSyntaxAnalyzer


class RWPMManifestParserFactory(ManifestParserFactory):
    """Factory creating RWPM parsers."""

    def create(self):
        """Create a new RWPMManifestParser.

        :return: RWPM parser instance
        :rtype: Parser
        """
        media_types_registry = RWPMMediaTypesRegistry()
        link_relations_registry = RWPMLinkRelationsRegistry()
        collection_roles_registry = RWPMCollectionRolesRegistry()
        syntax_analyzer = RWPMSyntaxAnalyzer()
        semantic_analyzer = RWPMSemanticAnalyzer(
            media_types_registry, link_relations_registry, collection_roles_registry
        )
        parser = ManifestParser(syntax_analyzer, semantic_analyzer)

        return parser



from webpub_manifest_parser.core import ManifestParser, ManifestParserFactory
from webpub_manifest_parser.opds2.registry import (
    OPDS2CollectionRolesRegistry,
    OPDS2LinkRelationsRegistry,
    OPDS2MediaTypesRegistry,
)
from webpub_manifest_parser.opds2.semantic import OPDS2SemanticAnalyzer
from webpub_manifest_parser.opds2.syntax import OPDS2SyntaxAnalyzer


class OPDS2FeedParserFactory(ManifestParserFactory):
    """Factory creating OPDS 2.0 parser."""

    def create(self):
        """Create a new OPDS 2.0 parser.

        :return: OPDS 2.0 parser
        :rtype: Parser
        """
        media_types_registry = OPDS2MediaTypesRegistry()
        link_relations_registry = OPDS2LinkRelationsRegistry()
        collection_roles_registry = OPDS2CollectionRolesRegistry()
        syntax_analyzer = OPDS2SyntaxAnalyzer()
        semantic_analyzer = OPDS2SemanticAnalyzer(
            media_types_registry, link_relations_registry, collection_roles_registry
        )
        parser = ManifestParser(syntax_analyzer, semantic_analyzer)

        return parser



from webpub_manifest_parser.core.syntax import SyntaxAnalyzer
from webpub_manifest_parser.rwpm.ast import RWPMManifest


class RWPMSyntaxAnalyzer(SyntaxAnalyzer):
    """Syntax analyzer for RWPM grammar."""

    def _create_manifest(self):
        """Create a new RWPM manifest.

        :return: RWPM manifest
        :rtype: RWPMManifest
        """
        return RWPMManifest()



import datetime

from webpub_manifest_parser.core.ast import (
    ArrayOfCollectionsProperty,
    ArrayOfLinksProperty,
    Collection,
    CollectionList,
    Node,
)
from webpub_manifest_parser.core.properties import (
    ArrayOfStringsProperty,
    BooleanProperty,
    DateOrTimeProperty,
    NumberProperty,
    TypeProperty,
    URIProperty,
)
from webpub_manifest_parser.odl.registry import ODLCollectionRolesRegistry
from webpub_manifest_parser.opds2.ast import OPDS2Feed, OPDS2Price, OPDS2Publication
from webpub_manifest_parser.opds2.registry import OPDS2CollectionRolesRegistry
from webpub_manifest_parser.utils import is_string


class ODLLicenseTerms(Node):
    """ODL license terms & conditions."""

    checkouts = NumberProperty("checkouts", required=False)
    expires = DateOrTimeProperty("expires", required=False)
    concurrency = NumberProperty("concurrency", required=False)
    length = NumberProperty("length", required=False)


class ODLLicenseProtection(Node):
    """ODL license protection information."""

    formats = ArrayOfStringsProperty("format", required=False)
    devices = NumberProperty("devices", required=False)
    copy_allowed = BooleanProperty("copy", required=False)
    print_allowed = BooleanProperty("print", required=False)
    tts_allowed = BooleanProperty("tts", required=False)


class ODLLicenseMetadata(Node):
    """ODL license metadata."""

    identifier = URIProperty("identifier", required=True)
    formats = ArrayOfStringsProperty("format", required=True)
    created = DateOrTimeProperty("created", required=True)
    terms = TypeProperty(
        "terms",
        required=False,
        nested_type=ODLLicenseTerms,
    )
    protection = TypeProperty(
        "protection",
        required=False,
        nested_type=ODLLicenseProtection,
    )
    price = TypeProperty(
        "price",
        required=False,
        nested_type=OPDS2Price,
    )
    source = URIProperty("source", required=False)

    def __init__(self, identifier=None, formats=None, created=None):
        """Initialize a new instance of ODLLicenseMetadata class.

        :param identifier: License's identifier
        :type identifier: str

        :param formats: List of license formats
        :type formats: List[str]

        :param created: Time when the license was created
        :type created: datetime.datetime
        """
        super(ODLLicenseMetadata, self).__init__()

        if identifier and not is_string(identifier):
            raise ValueError("Argument 'identifier' must be a string")
        if formats and not isinstance(formats, list):
            raise ValueError("Argument 'formats' must be a list")
        if created and not isinstance(created, datetime.datetime):
            raise ValueError(
                "Argument 'created' must be an instance of {0}".format(
                    datetime.datetime
                )
            )

        self.identifier = identifier
        self.formats = formats
        self.created = created


class ODLLicense(Collection):
    """ODL license subcollection."""

    metadata = TypeProperty("metadata", required=True, nested_type=ODLLicenseMetadata)

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash((self.metadata, self.links))


class ODLPublication(OPDS2Publication):
    """ODL publication."""

    links = ArrayOfLinksProperty(key="links", required=False)
    licenses = ArrayOfCollectionsProperty(
        "licenses",
        required=False,
        role=ODLCollectionRolesRegistry.LICENSES,
        collection_type=ODLLicense,
    )

    def __init__(self, metadata=None, links=None, images=None, licenses=None):
        """Initialize a new instance of ODLPublication class.

        :param metadata: Publication's metadata
        :type metadata: webpub_manifest_parser.core.ast.Metadata

        :param links: List of publication's links
        :type links: LinkList

        :param images: List of publication's images
        :type images: LinkList

        :param licenses: List of publication's licenses
        :type licenses: LinkList
        """
        super(ODLPublication, self).__init__(metadata, links, images)

        if licenses and not isinstance(licenses, CollectionList):
            raise ValueError(
                "Argument 'licenses' must be an instance of {0}".format(CollectionList)
            )

        self.licenses = licenses

    def __hash__(self):
        """Calculate the hash.

        :return: Hash
        :rtype: int
        """
        return hash((super(ODLPublication, self).__hash__(), self.licenses))


class ODLFeed(OPDS2Feed):
    """ODL 2.x feed."""

    publications = ArrayOfCollectionsProperty(
        "publications",
        required=False,
        role=OPDS2CollectionRolesRegistry.PUBLICATIONS,
        collection_type=ODLPublication,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['webpub_manifest_parser',
 'webpub_manifest_parser.core',
 'webpub_manifest_parser.epub',
 'webpub_manifest_parser.odl',
 'webpub_manifest_parser.opds2',
 'webpub_manifest_parser.rwpm']

package_data = \
{'': ['*']}

install_requires = \
['enum34>=1.1.10,<2.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'multipledispatch>=0.6.0,<0.7.0',
 'pyrsistent==0.16.1',
 'rfc3987>=1.3.8,<2.0.0',
 'strict-rfc3339>=0.7,<0.8',
 'uritemplate>=3.0.1,<4.0.0']

setup_kwargs = {
    'name': 'palace-webpub-manifest-parser',
    'version': '2.0.1',
    'description': 'A parser for the Readium Web Publication Manifest, OPDS 2.0 and ODL formats.',
    'long_description': '# webpub manifest parser\n\n[![Lint & Run Tests](https://github.com/ThePalaceProject/webpub-manifest-parser/actions/workflows/lint-test.yml/badge.svg)](https://github.com/ThePalaceProject/webpub-manifest-parser/actions/workflows/lint-test.yml)\n[![PyPI](https://img.shields.io/pypi/v/palace-webpub-manifest-parser)](https://pypi.org/project/palace-webpub-manifest-parser/)\n\nA parser for the [Readium Web Publication Manifest (RWPM)](https://github.com/readium/webpub-manifest), [Open Publication Distribution System 2.0 (OPDS 2.0)](https://drafts.opds.io/opds-2.0), and [Open Distribution to Libraries 1.0 (ODL)](https://drafts.opds.io/odl-1.0.html) formats.\n\n## Usage\nInstall the library with `pip`\n```bash\npip install palace-webpub-manifest-parser\n``` \n\n### Pyenv\n\nYou can optionally install the python version to run the library with using pyenv.\n\n1. Install [pyenv](https://github.com/pyenv/pyenv#installation)\n\n3. Install one of the supported Python versions:\n```bash\npyenv install <python-version>\n```\n\n4. Install [pyenv-virtualenv](https://github.com/pyenv/pyenv-virtualenv#installation) plugin\n\n5. Create a virtual environment:\n```bash\npyenv virtualenv <virtual-env-name>\npyenv activate <virtual-env-name>\n```\n\n6. Install the library\n```bash\npip install palace-webpub-manifest-parser\n``` \n\n\n# Setting up a development environment\n\n## Running tests using tox\n1. Make sure that a virtual environment is not activated and deactivate it if needed:\n```bash\ndeactivate\n```\n\n2. Install `tox` and `tox-pyenv` globally:\n```bash\npip install tox tox-pyenv\n```\n\n3. Make your code prettier using isort and black:\n```bash\nmake reformat\n``` \n\n4. Run the linters:\n```bash\nmake lint\n```\n\n5. To run the unit tests use the following command:\n```bash\nmake test-<python-version>\n```\nwhere `<python-version>` is one of supported python versions:\n- py27\n- py36\n- py37\n- py38\n\nFor example, to run the unit test using Python 2.7 run the following command:\n```bash\nmake test-py27\n```\n\n# Releasing\n\nReleases will be automatically published to PyPI when new releases are created on github. We use \n`bump2version` to update the version number, then create a release in github.\n\nTo publish a new release:\n- Bump the version and push it as a branch\n  ```\n  git checkout -b release/vX.X.X\n  pip install bump2version\n  bump2version {part}\n  git push origin release/vX.X.X\n  ```\n- Create a PR for new version\n- Merge PR into `main`\n- Create a release for the version in github\n',
    'author': 'Viacheslav Bessonov',
    'author_email': 'viacheslav.bessonov@hilbertteam.com',
    'maintainer': 'Viacheslav Bessonov',
    'maintainer_email': 'viacheslav.bessonov@hilbertteam.com',
    'url': 'https://github.com/ThePalaceProject/python-webpub-manifest-parser',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
}


setup(**setup_kwargs)



import collections


class RegistryItem(object):
    """Single metadata registry item (collection role, media type, etc.)."""

    def __init__(self, key):
        """Initialize a new instance of RegistryItem class.

        :param key: Unique identifier of this registry item
        :type key: str
        """
        self._key = key

    @property
    def key(self):
        """Return a unique identifier of this registry item.

        :return: Unique identifier of this registry item
        :rtype: str
        """
        return self._key


class Registry(collections.MutableMapping):
    """Collection of registry items with a particular type (collection roles, media types, etc.)."""

    def __init__(self, items=None):
        """Initialize a new instance of Registry class.

        :param items: (Optional) collection of registry items. Note that all items have to be RegistryItem descendants
        :type items: List[RegistryItems]
        """
        self._items = {}

        if items:
            self._add_items(items)

    def __setitem__(self, key, value):
        """Add a new item to the registry.

        :param key: Unique identifier of the item
        :type key: str

        :param value: Registry item
        :type value: RegistryItem
        """
        if not isinstance(value, RegistryItem):
            raise ValueError("Registry item must have RegistryItem type")

        self._items[key] = value

    def __delitem__(self, key):
        """Remove the item from the registry.

        :param key: Unique identifier of the item
        :type key: str
        """
        del self._items[key]

    def __getitem__(self, key):
        """Return an item from the registry by its key or raises a KeyError if it doesn't exist.

        :param key: Unique identifier of the item
        :type key: str

        :return: Registry item
        :rtype: RegistryItem
        """
        return self._items[key]

    def __iter__(self):
        """Return an iterator of all the registry items.

        :return: Iterator of the registry items
        :rtype: Iterator[RegistryItem]
        """
        return iter(self._items.values())

    def __len__(self):
        """Return a number of items in the registry.

        :return: Number of items in the registry
        :rtype: int
        """
        return len(self._items)

    def _add_items(self, items):
        """Add new items to the registry. Note that all the items must be RegistryItem descendants.

        :param items: New registry items
        :type items: List[RegistryItem]
        """
        for item in items:
            if not isinstance(item, RegistryItem):
                raise ValueError("Registry item must have RegistryItem type")

            self._items[item.key] = item


class MediaType(RegistryItem):
    """Registry item representing a specific media type."""


class LinkRelation(RegistryItem):
    """Registry item representing a link relation."""


class CollectionRole(RegistryItem):
    """Registry item representing a collection role."""

    def __init__(self, key, compact, required, multi=False):
        """Initialize a new instance of CollectionRole class.

        :param key: Name of the collection
        :type key: str

        :param compact: Boolean value indicating whether the collection shall be compact
        :type compact: bool

        :param required: Boolean value indicating whether the collection is required
        :type required: bool

        :param multi: Boolean value indicating whether there can be multiple collections with this role
        :type multi: bool
        """
        super(CollectionRole, self).__init__(key)

        self._compact = compact
        self._required = required
        self._multi = multi

    @property
    def compact(self):
        """Return the boolean value indicating whether the collection shall be compact.

        :return: Boolean value indicating whether the collection shall be compact
        :rtype: bool
        """
        return self._compact

    @property
    def required(self):
        """Return the boolean value indicating whether the collection is required.

        :return: Boolean value indicating whether the collection is required
        :rtype: bool
        """
        return self._required

    @property
    def multi(self):
        """Return the boolean value indicating whether there can be multiple collections with this role.

        :return: Boolean value indicating whether there can be multiple collections with this role
        :rtype: bool
        """
        return self._multi


class LinkRelationsRegistry(Registry):
    """Registry containing link relations mentioned in the RWPM spec."""

    SELF = LinkRelation(key="self")

    CORE_LINK_RELATIONS = [SELF]

    def __init__(self, items=None):
        """Initialize a new instance of LinkRelationsRegistry class.

        :param items: (Optional) collection of registry items. Note that all items have to be RegistryItem descendants
        :type items: List[RegistryItems]
        """
        super(LinkRelationsRegistry, self).__init__(
            self.CORE_LINK_RELATIONS if not items else items
        )
